#include <stdio.h>
#include "traceplay.h"
#include "common.h"

const char * cuda_test_name = NULL;

#define NUM_OUTPUT_SURFACES 2

struct Traceplay_Ctx{
    VdpDevice                 vdp_device;
    VdpGetProcAddress *       vdp_get_proc_address;
    Traceplay_Play_Type       play_type;
    Traceplay_Callbacks *     callbacks;
    VdpDecoderProfile         decoder_profile;
    VdpGetErrorString *       vdp_get_error_string;
    VdpVideoSurfaceCreate *   vdp_video_surface_create;
    VdpVideoSurfaceDestroy *  vdp_video_surface_destroy;
    VdpOutputSurfaceCreate *  vdp_output_surface_create;
    VdpOutputSurfaceDestroy * vdp_output_surface_destroy;
    VdpDecoderCreate *        vdp_decoder_create;
    VdpDecoderDestroy *       vdp_decoder_destroy;
    VdpDecoderRender *        vdp_decoder_render;
    VdpVideoMixerCreate *     vdp_video_mixer_create;
    VdpVideoMixerDestroy *    vdp_video_mixer_destroy;
    VdpVideoMixerRender *     vdp_video_mixer_render;
    VdpDecoder                decoder;
    VdpVideoMixer             mixer;
    uint32_t                  vid_surface_count;
    VdpVideoSurface *         vid_surfaces;
    void * *                  vid_surface_callback_contexts;
    uint32_t                  out_surface_count;
    VdpOutputSurface *        out_surfaces;
    void * *                  out_surface_callback_contexts;
    int                       output_index;
};

struct Traceplay_Ctx * traceplay_init(
    VdpDevice               vdp_device,
    VdpGetProcAddress *     vdp_get_proc_address,
    Traceplay_Play_Type     play_type,
    Traceplay_Callbacks *   callbacks,
    Traceplay_Stream_Conf * stream_conf
)
{
    struct Traceplay_Ctx * ctx;
    int i;
    VdpStatus vdp_st;

    ctx = calloc(1, sizeof(struct Traceplay_Ctx));

    ctx->vdp_device = vdp_device;
    ctx->vdp_get_proc_address = vdp_get_proc_address;
    ctx->play_type = play_type;
    ctx->callbacks = callbacks;

    ctx->decoder_profile = stream_conf->decoder_profile;

    ctx->decoder = VDP_INVALID_HANDLE;
    ctx->mixer = VDP_INVALID_HANDLE;

    ctx->vid_surface_count = stream_conf->vid_surface_count;
    ctx->vid_surfaces = malloc(stream_conf->vid_surface_count * sizeof(VdpVideoSurface));
    if (!ctx->vid_surfaces) {
        printf("Error at %s:%d: !ctx->vid_surfaces\n", __FILE__, __LINE__);
        common_exit_error();
    }
    for (i = 0; i < ctx->vid_surface_count; ++i) {
        ctx->vid_surfaces[i] = VDP_INVALID_HANDLE;
    }
    ctx->vid_surface_callback_contexts = malloc(stream_conf->vid_surface_count * sizeof(void*));
    if (!ctx->vid_surface_callback_contexts) {
        printf("Error at %s:%d: !ctx->vid_surface_callback_contexts\n", __FILE__, __LINE__);
        common_exit_error();
    }

    if (ctx->play_type == TRACEPLAY_PLAY_TYPE_VIDEO) {
        ctx->out_surfaces = 0;
    }
    else {
        ctx->out_surface_count = NUM_OUTPUT_SURFACES;
        ctx->out_surfaces = malloc(ctx->out_surface_count);
        if (!ctx->out_surfaces) {
            printf("Error at %s:%d: !ctx->out_surfaces\n", __FILE__, __LINE__);
            common_exit_error();
        }
        for (i = 0; i < ctx->out_surface_count; ++i) {
            ctx->out_surfaces[i] = VDP_INVALID_HANDLE;
        }
        ctx->out_surface_callback_contexts = malloc(ctx->out_surface_count * sizeof(void*));
        if (!ctx->out_surface_callback_contexts) {
            printf("Error at %s:%d: !ctx->out_surface_callback_contexts\n", __FILE__, __LINE__);
            common_exit_error();
        }
    }

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_VIDEO_SURFACE_CREATE,
        (void **)&ctx->vdp_video_surface_create
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_VIDEO_SURFACE_DESTROY,
        (void **)&ctx->vdp_video_surface_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_OUTPUT_SURFACE_CREATE,
        (void **)&ctx->vdp_output_surface_create
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_OUTPUT_SURFACE_DESTROY,
        (void **)&ctx->vdp_output_surface_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_DECODER_CREATE,
        (void **)&ctx->vdp_decoder_create
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_DECODER_DESTROY,
        (void **)&ctx->vdp_decoder_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_DECODER_RENDER,
        (void **)&ctx->vdp_decoder_render
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_VIDEO_MIXER_CREATE,
        (void **)&ctx->vdp_video_mixer_create
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_VIDEO_MIXER_DESTROY,
        (void **)&ctx->vdp_video_mixer_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_VIDEO_MIXER_RENDER,
        (void **)&ctx->vdp_video_mixer_render
    );
    CHECK_ST

    vdp_st = ctx->vdp_decoder_create(
        ctx->vdp_device,
        stream_conf->decoder_profile,
        stream_conf->vid_width,
        stream_conf->vid_height,
        stream_conf->decoder_max_references,
        &(ctx->decoder)
    );
    CHECK_ST

    if (ctx->play_type == TRACEPLAY_PLAY_TYPE_OUTPUT) {
        VdpChromaType chroma_type = VDP_CHROMA_TYPE_420;
        VdpVideoMixerParameter parameters[] = {
            VDP_VIDEO_MIXER_PARAMETER_VIDEO_SURFACE_WIDTH,
            VDP_VIDEO_MIXER_PARAMETER_VIDEO_SURFACE_HEIGHT,
            VDP_VIDEO_MIXER_PARAMETER_CHROMA_TYPE
        };
        void const * parameter_values[] = {
            &stream_conf->vid_width,
            &stream_conf->vid_height,
            &chroma_type
        };

        vdp_st = ctx->vdp_video_mixer_create(
            ctx->vdp_device,
            0,
            NULL,
            3,
            parameters,
            parameter_values,
            &(ctx->mixer)
        );
        CHECK_ST
    }

    for (i = 0; i < ctx->vid_surface_count; ++i) {
        vdp_st = ctx->vdp_video_surface_create(
            ctx->vdp_device,
            VDP_CHROMA_TYPE_420,
            stream_conf->vid_width,
            stream_conf->vid_height,
            &(ctx->vid_surfaces[i])
        );
        CHECK_ST
        ctx->vid_surface_callback_contexts[i] =
            ctx->callbacks->video_surface_post_create(ctx, ctx->vid_surfaces[i]);
    }

    for (i = 0; i < ctx->out_surface_count; ++i) {
        vdp_st = ctx->vdp_output_surface_create(
            ctx->vdp_device,
            VDP_RGBA_FORMAT_B8G8R8A8,
            stream_conf->out_width,
            stream_conf->out_height,
            &(ctx->out_surfaces[i])
        );
        CHECK_ST
        ctx->out_surface_callback_contexts[i] =
            ctx->callbacks->output_surface_post_create(ctx, ctx->out_surfaces[i]);
    }

    return ctx;
}

void traceplay_fini(
    struct Traceplay_Ctx * ctx
)
{
    int i;

    for (i = 0; i < ctx->out_surface_count; ++i) {
        if (ctx->out_surfaces[i] == VDP_INVALID_HANDLE) {
            continue;
        }

        ctx->callbacks->output_surface_pre_destroy(
            ctx,
            ctx->out_surfaces[i],
            ctx->out_surface_callback_contexts[i]
        );
        ctx->vdp_output_surface_destroy(
            ctx->out_surfaces[i]
        );
    }

    for (i = 0; i < ctx->vid_surface_count; ++i) {
        if (ctx->vid_surfaces[i] == VDP_INVALID_HANDLE) {
            continue;
        }

        ctx->callbacks->video_surface_pre_destroy(
            ctx,
            ctx->vid_surfaces[i],
            ctx->vid_surface_callback_contexts[i]
        );
        ctx->vdp_video_surface_destroy(
            ctx->vid_surfaces[i]
        );
    }

    if (ctx->mixer != VDP_INVALID_HANDLE) {
        ctx->vdp_video_mixer_destroy(
            ctx->mixer
        );
    }

    if (ctx->decoder != VDP_INVALID_HANDLE) {
        ctx->vdp_decoder_destroy(
            ctx->decoder
        );
    }

    if (ctx->out_surface_callback_contexts) {
        free(ctx->out_surface_callback_contexts);
    }

    if (ctx->out_surfaces) {
        free(ctx->out_surfaces);
    }

    if (ctx->vid_surface_callback_contexts) {
        free(ctx->vid_surface_callback_contexts);
    }

    if (ctx->vid_surfaces) {
        free(ctx->vid_surfaces);
    }

    free(ctx);
}

void traceplay_get_vid_surfaces(
    struct Traceplay_Ctx * ctx,
    uint32_t *             surface_count,
    VdpVideoSurface * *    surfaces
)
{
    *surface_count = ctx->vid_surface_count;
    *surfaces = ctx->vid_surfaces;
    
}

void traceplay_get_out_surfaces(
    struct Traceplay_Ctx * ctx,
    uint32_t *             surface_count,
    VdpOutputSurface * *   surfaces
)
{
    *surface_count = ctx->out_surface_count;
    *surfaces = ctx->out_surfaces;
}

void traceplay_play(
    struct Traceplay_Ctx * ctx,
    Traceplay_Cmds *       cmds,
    int                    repeat
)
{
    int i, f, iter;
    uint32_t dec_frame_number = 0, disp_frame_number = 0;

    for (iter = 0; (!repeat) || (iter < repeat); iter++) {
        int done = 0;

        for (i = 0; !done; ++i) {
            switch (cmds[i].cmd) {
            case TRACEPLAY_CMD_DECODE:
                {
                    Traceplay_CmdDecode * cmd = (Traceplay_CmdDecode *)cmds[i].data;
                    VdpPictureInfoMPEG1Or2 pimpeg;
                    VdpPictureInfoVC1 pivc1;
                    VdpPictureInfoH264 pih264;
                    VdpPictureInfoMPEG4Part2 pimp4p2;
                    VdpPictureInfo pi = 0;

                    switch (ctx->decoder_profile) {
                    case VDP_DECODER_PROFILE_MPEG1:
                    case VDP_DECODER_PROFILE_MPEG2_SIMPLE:
                    case VDP_DECODER_PROFILE_MPEG2_MAIN:
                        pimpeg = *(VdpPictureInfoMPEG1Or2 *)cmd->picture_info;
                        if (pimpeg.forward_reference != VDP_INVALID_HANDLE) {
                            pimpeg.forward_reference = ctx->vid_surfaces[pimpeg.forward_reference];
                        }
                        if (pimpeg.backward_reference != VDP_INVALID_HANDLE) {
                            pimpeg.backward_reference = ctx->vid_surfaces[pimpeg.backward_reference];
                        }
                        pi = &pimpeg;
                        break;
                    case VDP_DECODER_PROFILE_VC1_SIMPLE:
                    case VDP_DECODER_PROFILE_VC1_MAIN:
                    case VDP_DECODER_PROFILE_VC1_ADVANCED:
                        pivc1 = *(VdpPictureInfoVC1 *)cmd->picture_info;
                        if (pivc1.forward_reference != VDP_INVALID_HANDLE) {
                            pivc1.forward_reference = ctx->vid_surfaces[pivc1.forward_reference];
                        }
                        if (pivc1.backward_reference != VDP_INVALID_HANDLE) {
                            pivc1.backward_reference = ctx->vid_surfaces[pivc1.backward_reference];
                        }
                        pi = &pivc1;
                        break;
                    case VDP_DECODER_PROFILE_H264_BASELINE:
                    case VDP_DECODER_PROFILE_H264_MAIN:
                    case VDP_DECODER_PROFILE_H264_HIGH:
                        pih264 = *(VdpPictureInfoH264 *)cmd->picture_info;
                        for (f = 0; f < 16; ++f) {
                            if (pih264.referenceFrames[f].surface != VDP_INVALID_HANDLE) {
                                pih264.referenceFrames[f].surface = ctx->vid_surfaces[pih264.referenceFrames[f].surface];
                            }
                        }
                        pi = &pih264;
                        break;
                    case VDP_DECODER_PROFILE_MPEG4_PART2_SP:
                    case VDP_DECODER_PROFILE_MPEG4_PART2_ASP:
                    case VDP_DECODER_PROFILE_DIVX4_QMOBILE:
                    case VDP_DECODER_PROFILE_DIVX4_MOBILE:
                    case VDP_DECODER_PROFILE_DIVX4_HOME_THEATER:
                    case VDP_DECODER_PROFILE_DIVX4_HD_1080P:
                    case VDP_DECODER_PROFILE_DIVX5_QMOBILE:
                    case VDP_DECODER_PROFILE_DIVX5_MOBILE:
                    case VDP_DECODER_PROFILE_DIVX5_HOME_THEATER:
                    case VDP_DECODER_PROFILE_DIVX5_HD_1080P:
                        pimp4p2 = *(VdpPictureInfoMPEG4Part2 *)cmd->picture_info;
                        if (pimp4p2.forward_reference != VDP_INVALID_HANDLE) {
                            pimp4p2.forward_reference = ctx->vid_surfaces[pimp4p2.forward_reference];
                        }
                        if (pimp4p2.backward_reference != VDP_INVALID_HANDLE) {
                            pimp4p2.backward_reference = ctx->vid_surfaces[pimp4p2.backward_reference];
                        }
                        pi = &pimp4p2;
                        break;
                    default:
                        printf("Error at %s:%d: Unsupported decoder profile\n", __FILE__, __LINE__);
                        common_exit_error();
                    }
                    
                    ctx->vdp_decoder_render(
                        ctx->decoder,
                        ctx->vid_surfaces[cmd->vid_surface_index],
                        pi,
                        cmd->bitstream_buffer_count,
                        cmd->bitstream_buffers
                    );
                    if (opt_gold_surface_type == SURFACE_TYPE_VDP_VID_DEC_ORDER) {
                        common_do_gold_vdp_video_surface(dec_frame_number, ctx->vid_surfaces[cmd->vid_surface_index]);
                    }
                    dec_frame_number++;
                }
                break;
            case TRACEPLAY_CMD_DISPLAY:
                {
                    Traceplay_CmdDisplay * cmd = (Traceplay_CmdDisplay *)cmds[i].data;

                    if (opt_gold_surface_type == SURFACE_TYPE_VDP_VID_DISP_ORDER) {
                        common_do_gold_vdp_video_surface(disp_frame_number, ctx->vid_surfaces[cmd->vid_surface_index]);
                    }

                    switch (ctx->play_type) {
                    case TRACEPLAY_PLAY_TYPE_VIDEO:
                        ctx->callbacks->display_video_surface(
                            ctx,
                            disp_frame_number,
                            ctx->vid_surfaces[cmd->vid_surface_index],
                            ctx->vid_surface_callback_contexts[cmd->vid_surface_index]
                        );
                        break;
                    case TRACEPLAY_PLAY_TYPE_OUTPUT:
                    case TRACEPLAY_PLAY_TYPE_CUSTOM_MIXER:
                        {
                            VdpOutputSurface os = ctx->out_surfaces[ctx->output_index];
                            void * osctx = ctx->out_surface_callback_contexts[ctx->output_index];
                            ctx->callbacks->wait_for_output_surface_idle(ctx, os, osctx);
                            if (ctx->play_type == TRACEPLAY_PLAY_TYPE_OUTPUT) {
                                ctx->vdp_video_mixer_render(
                                    ctx->mixer,
                                    VDP_INVALID_HANDLE,
                                    NULL,
                                    VDP_VIDEO_MIXER_PICTURE_STRUCTURE_FRAME,
                                    0,
                                    NULL,
                                    ctx->vid_surfaces[cmd->vid_surface_index],
                                    0,
                                    NULL,
                                    NULL,
                                    os,
                                    NULL,
                                    NULL,
                                    0,
                                    NULL
                                );
                            }
                            else {
                                ctx->callbacks->process_surface(
                                    ctx,
                                    disp_frame_number,
                                    ctx->vid_surfaces[cmd->vid_surface_index],
                                    ctx->vid_surface_callback_contexts[cmd->vid_surface_index],
                                    os,
                                    osctx
                                );
                            }
                            if (opt_gold_surface_type == SURFACE_TYPE_VDP_OUT) {
                                common_do_gold_vdp_output_surface(disp_frame_number, os);
                            }
                            ctx->callbacks->display_output_surface(ctx, disp_frame_number, os, osctx);
                            ctx->output_index = (ctx->output_index + 1) % ctx->out_surface_count;
                        }
                        break;
                    default:
                        break;
                    }
                    disp_frame_number++;
                }
                break;
            case TRACEPLAY_CMD_END:
                done = 1;
                break;
            }
        }
    }
}

