#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vdpau/vdpau_x11.h>
#include <cuda/cuda.h>
#include <cuda/cudaVDPAU.h>
#include "traceplay.h"
#include "common.h"

#define ALIGN_OFFSET(offset, alignment) offset = (offset + (alignment-1)) & ~((alignment-1))

typedef struct {
    CUgraphicsResource gr;
} RegisteredVideoSurface;

typedef struct {
    CUgraphicsResource gr;
} RegisteredOutputSurface;

Display *     x_display;
int           x_screen;
Window        x_window;

VdpDevice           vdp_device;
VdpGetProcAddress * vdp_get_proc_address;

VdpPresentationQueueTarget vdp_flip_target;
VdpPresentationQueue       vdp_flip_queue;

VdpDeviceDestroy *                          vdp_device_destroy;
VdpPresentationQueueTargetDestroy *         vdp_presentation_queue_target_destroy;
VdpPresentationQueueCreate *                vdp_presentation_queue_create;
VdpPresentationQueueDestroy *               vdp_presentation_queue_destroy;
VdpPresentationQueueDisplay *               vdp_presentation_queue_display;
VdpPresentationQueueBlockUntilSurfaceIdle * vdp_presentation_queue_block_until_surface_idle;
VdpPresentationQueueTargetCreateX11 *       vdp_presentation_queue_target_create_x11;

CUdevice cu_dev;
CUcontext cu_ctx;
CUmodule cu_mod_vmr;
CUfunction cu_func_luma_weave, cu_func_chroma_weave, cu_func_csc;
CUdeviceptr cu_dptr_out, cu_dptr_weaved_luma, cu_dptr_weaved_chroma;
unsigned int pitch_out, pitch_weaved_luma, pitch_weaved_chroma;
CUtexref cu_tr_luma_top, cu_tr_luma_bottom, cu_tr_chroma_top, cu_tr_chroma_bottom, cu_tr_luma, cu_tr_chroma;

void * traceplay_video_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface
)
{
    RegisteredVideoSurface *rvs = (RegisteredVideoSurface *)malloc(sizeof(*rvs));
    CHECK_PTR(rvs)
    
    CUresult cr = cuGraphicsVDPAURegisterVideoSurface(&rvs->gr, surface, CU_GRAPHICS_MAP_RESOURCE_FLAGS_READ_ONLY);
    CHECK_CU_RESULT

    return rvs;
}

void * traceplay_output_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface
)
{
    RegisteredOutputSurface *ros = (RegisteredOutputSurface *)malloc(sizeof(*ros));
    CHECK_PTR(ros)
    
    CUresult cr = cuGraphicsVDPAURegisterOutputSurface(&ros->gr, surface, CU_GRAPHICS_MAP_RESOURCE_FLAGS_WRITE_DISCARD);
    CHECK_CU_RESULT

    return ros;
}

void traceplay_video_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
    RegisteredVideoSurface *rvs = (RegisteredVideoSurface *)surface_context;

    CUresult cr = cuGraphicsUnregisterResource(rvs->gr);
    CHECK_CU_RESULT

    free(rvs);
}

void traceplay_output_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    RegisteredOutputSurface *ros = (RegisteredOutputSurface *)surface_context;

    CUresult cr = cuGraphicsUnregisterResource(ros->gr);
    CHECK_CU_RESULT

    free(ros);
}

void traceplay_display_video_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
    printf("ERROR at %s:%d: display_video_surface called\n", __FILE__, __LINE__);
    common_exit_error();
}

void traceplay_wait_for_output_surface_idle(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    VdpTime t;
    vdp_presentation_queue_block_until_surface_idle(vdp_flip_queue, surface, &t);
}

void traceplay_display_output_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    vdp_presentation_queue_display(vdp_flip_queue, surface, 0, 0, 0);
}

void cuda_driver_do_gold_interop_vid(uint32_t frame_number, CUarray arvtl, CUarray arvbl, CUarray arvtc, CUarray arvbc)
{
    int i;
    CUresult cr;
    CUDA_MEMCPY2D copy;

    if (opt_gold_every && (frame_number % opt_gold_every)) {
        return;
    }

    uint8_t *buffer = scratch_surface_buffer;
    size_t height = trace_stream_conf_ptr->vid_height / 2;
    size_t texsize = trace_stream_conf_ptr->vid_width * height;
    CUarray ars[] = {arvtl, arvbl, arvtc, arvbc};

    for (i = 0; i < 4; ++i) {
        memset(&copy, 0, sizeof(copy));
        copy.srcXInBytes = 0;
        copy.srcY = 0;
        copy.srcMemoryType = CU_MEMORYTYPE_ARRAY;
        copy.srcArray = ars[i];
        copy.dstXInBytes = 0;
        copy.dstY = 0;
        copy.dstMemoryType = CU_MEMORYTYPE_HOST;
        copy.dstHost = buffer;
        copy.dstPitch = trace_stream_conf_ptr->vid_width;
        copy.WidthInBytes = trace_stream_conf_ptr->vid_width;
        copy.Height = height;
        cr = cuMemcpy2D(&copy);
        CHECK_CU_RESULT

        buffer += texsize;
        if (i == 1) {
            height /= 2;
            texsize /= 2;
        }
    }

    common_nv12x2_to_nv12(trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height, scratch_surface_buffer, surface_buffer);

    common_do_gold_buffer(frame_number, "nv12", trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height, surface_buffer, surface_buffer_size);
}

void cuda_driver_do_gold_interop_out(uint32_t frame_number, CUarray aro)
{
    CUresult cr;
    CUDA_MEMCPY2D copy;

    if (opt_gold_every && (frame_number % opt_gold_every)) {
        return;
    }

    memset(&copy, 0, sizeof(copy));
    copy.srcXInBytes = 0;
    copy.srcY = 0;
    copy.srcMemoryType = CU_MEMORYTYPE_ARRAY;
    copy.srcArray = aro;
    copy.dstXInBytes = 0;
    copy.dstY = 0;
    copy.dstMemoryType = CU_MEMORYTYPE_HOST;
    copy.dstHost = surface_buffer;
    copy.dstPitch = trace_stream_conf_ptr->out_width * 4;
    copy.WidthInBytes = trace_stream_conf_ptr->out_width * 4;
    copy.Height = trace_stream_conf_ptr->out_height;
    cr = cuMemcpy2D(&copy);
    CHECK_CU_RESULT

    common_do_gold_buffer_argb_to_rgb(frame_number, trace_stream_conf_ptr->out_width, trace_stream_conf_ptr->out_height, surface_buffer, surface_buffer_size, 0);
}

void traceplay_process_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        vid_surface,
    void *                 vid_surface_context,
    VdpOutputSurface       out_surface,
    void *                 out_surface_context
)
{
    RegisteredVideoSurface *rvs = (RegisteredVideoSurface *)vid_surface_context;
    RegisteredOutputSurface *ros = (RegisteredOutputSurface *)out_surface_context;
    CUresult cr;
    int offset;
    void *dptr;
    unsigned int w, h;
    CUDA_ARRAY_DESCRIPTOR desc;
    CUDA_MEMCPY2D copy;

    // Map and acuire arrays

    cr = cuGraphicsMapResources(1, &rvs->gr, 0);
    CHECK_CU_RESULT
    cr = cuGraphicsMapResources(1, &ros->gr, 0);
    CHECK_CU_RESULT

    CUarray arvtl, arvbl, arvtc, arvbc, aro;

    cr = cuGraphicsSubResourceGetMappedArray(&arvtl, rvs->gr, 0, 0);
    CHECK_CU_RESULT
    cr = cuGraphicsSubResourceGetMappedArray(&arvbl, rvs->gr, 1, 0);
    CHECK_CU_RESULT
    cr = cuGraphicsSubResourceGetMappedArray(&arvtc, rvs->gr, 2, 0);
    CHECK_CU_RESULT
    cr = cuGraphicsSubResourceGetMappedArray(&arvbc, rvs->gr, 3, 0);
    CHECK_CU_RESULT
    cr = cuGraphicsSubResourceGetMappedArray(&aro, ros->gr, 0, 0);
    CHECK_CU_RESULT

    if (opt_gold_surface_type == SURFACE_TYPE_CUDA_DRIVER_INTEROP_VID) {
        cuda_driver_do_gold_interop_vid(frame_number, arvtl, arvbl, arvtc, arvbc);
    }

    // Weave luma

    offset = 0;
    dptr = (void *)(size_t)cu_dptr_weaved_luma;
    ALIGN_OFFSET(offset, __alignof(dptr)); 
    cr = cuParamSetv(cu_func_luma_weave, offset, &dptr, sizeof(dptr));
    CHECK_CU_RESULT
    offset += sizeof(dptr);
    ALIGN_OFFSET(offset, __alignof(pitch_weaved_luma)); 
    cr = cuParamSetv(cu_func_luma_weave, offset, &pitch_weaved_luma, sizeof(pitch_weaved_luma));
    CHECK_CU_RESULT
    offset += sizeof(pitch_weaved_luma);
    cr = cuParamSetSize(cu_func_luma_weave, offset);
    CHECK_CU_RESULT
    cr = cuTexRefSetArray(cu_tr_luma_top, arvtl, CU_TRSA_OVERRIDE_FORMAT);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_luma_top, 0, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_luma_top, 1, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetFilterMode(cu_tr_luma_top, CU_TR_FILTER_MODE_POINT);
    CHECK_CU_RESULT
    cr = cuTexRefSetFlags(cu_tr_luma_top, CU_TRSF_READ_AS_INTEGER);
    CHECK_CU_RESULT
    cr = cuParamSetTexRef(cu_func_luma_weave, CU_PARAM_TR_DEFAULT, cu_tr_luma_top);
    CHECK_CU_RESULT
    cr = cuTexRefSetArray(cu_tr_luma_bottom, arvbl, CU_TRSA_OVERRIDE_FORMAT);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_luma_bottom, 0, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_luma_bottom, 1, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetFilterMode(cu_tr_luma_bottom, CU_TR_FILTER_MODE_POINT);
    CHECK_CU_RESULT
    cr = cuTexRefSetFlags(cu_tr_luma_bottom, CU_TRSF_READ_AS_INTEGER);
    CHECK_CU_RESULT
    cr = cuParamSetTexRef(cu_func_luma_weave, CU_PARAM_TR_DEFAULT, cu_tr_luma_bottom);
    CHECK_CU_RESULT
    cr = cuFuncSetBlockShape(cu_func_luma_weave, 16, 16, 1);
    CHECK_CU_RESULT
    cr = cuLaunchGrid(cu_func_luma_weave, trace_stream_conf_ptr->vid_width / 16, trace_stream_conf_ptr->vid_height / 16);
    CHECK_CU_RESULT

    // Weave Chroma

    offset = 0;
    dptr = (void *)(size_t)cu_dptr_weaved_chroma;
    ALIGN_OFFSET(offset, __alignof(dptr)); 
    cr = cuParamSetv(cu_func_chroma_weave, offset, &dptr, sizeof(dptr));
    CHECK_CU_RESULT
    offset += sizeof(dptr);
    ALIGN_OFFSET(offset, __alignof(pitch_weaved_chroma)); 
    cr = cuParamSetv(cu_func_chroma_weave, offset, &pitch_weaved_chroma, sizeof(pitch_weaved_chroma));
    CHECK_CU_RESULT
    offset += sizeof(pitch_weaved_chroma);
    cr = cuParamSetSize(cu_func_chroma_weave, offset);
    CHECK_CU_RESULT
    cr = cuTexRefSetArray(cu_tr_chroma_top, arvtc, CU_TRSA_OVERRIDE_FORMAT);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_chroma_top, 0, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_chroma_top, 1, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetFilterMode(cu_tr_chroma_top, CU_TR_FILTER_MODE_POINT);
    CHECK_CU_RESULT
    cr = cuTexRefSetFlags(cu_tr_chroma_top, CU_TRSF_READ_AS_INTEGER);
    CHECK_CU_RESULT
    cr = cuParamSetTexRef(cu_func_chroma_weave, CU_PARAM_TR_DEFAULT, cu_tr_chroma_top);
    CHECK_CU_RESULT
    cr = cuTexRefSetArray(cu_tr_chroma_bottom, arvbc, CU_TRSA_OVERRIDE_FORMAT);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_chroma_bottom, 0, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_chroma_bottom, 1, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetFilterMode(cu_tr_chroma_bottom, CU_TR_FILTER_MODE_POINT);
    CHECK_CU_RESULT
    cr = cuTexRefSetFlags(cu_tr_chroma_bottom, CU_TRSF_READ_AS_INTEGER);
    CHECK_CU_RESULT
    cr = cuParamSetTexRef(cu_func_chroma_weave, CU_PARAM_TR_DEFAULT, cu_tr_chroma_bottom);
    CHECK_CU_RESULT
    cr = cuFuncSetBlockShape(cu_func_chroma_weave, 8, 8, 1);
    CHECK_CU_RESULT
    cr = cuLaunchGrid(cu_func_chroma_weave, trace_stream_conf_ptr->vid_width / (8 * 2), trace_stream_conf_ptr->vid_height / (8 * 2));
    CHECK_CU_RESULT

    // CSC

    offset = 0;

    dptr = (void *)(size_t)cu_dptr_out;
    ALIGN_OFFSET(offset, __alignof(dptr)); 
    cr = cuParamSetv(cu_func_csc, offset, &dptr, sizeof(dptr));
    CHECK_CU_RESULT
    offset += sizeof(dptr);
    ALIGN_OFFSET(offset, __alignof(pitch_out)); 
    cr = cuParamSetv(cu_func_csc, offset, &pitch_out, sizeof(pitch_out));
    CHECK_CU_RESULT
    offset += sizeof(pitch_out);
    w = trace_stream_conf_ptr->out_width;
    ALIGN_OFFSET(offset, __alignof(w)); 
    cr = cuParamSetv(cu_func_csc, offset, &w, sizeof(w));
    CHECK_CU_RESULT
    offset += sizeof(w);
    h = trace_stream_conf_ptr->out_height;
    ALIGN_OFFSET(offset, __alignof(h)); 
    cr = cuParamSetv(cu_func_csc, offset, &h, sizeof(h));
    CHECK_CU_RESULT
    offset += sizeof(h);
    cr = cuParamSetSize(cu_func_csc, offset);
    CHECK_CU_RESULT
    desc.Width = trace_stream_conf_ptr->vid_width;
    desc.Height = trace_stream_conf_ptr->vid_height;
    desc.Format = CU_AD_FORMAT_UNSIGNED_INT8;
    desc.NumChannels = 1;
    cr = cuTexRefSetAddress2D(cu_tr_luma, &desc, cu_dptr_weaved_luma, pitch_weaved_luma);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_luma, 0, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_luma, 1, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetFilterMode(cu_tr_luma, CU_TR_FILTER_MODE_LINEAR);
    CHECK_CU_RESULT
    cr = cuTexRefSetFlags(cu_tr_luma, CU_TRSF_NORMALIZED_COORDINATES);
    CHECK_CU_RESULT
    cr = cuParamSetTexRef(cu_func_csc, CU_PARAM_TR_DEFAULT, cu_tr_luma);
    CHECK_CU_RESULT
    desc.Width = trace_stream_conf_ptr->vid_width / 2;
    desc.Height = trace_stream_conf_ptr->vid_height / 2;
    desc.Format = CU_AD_FORMAT_UNSIGNED_INT8;
    desc.NumChannels = 2;
    cr = cuTexRefSetAddress2D(cu_tr_chroma, &desc, cu_dptr_weaved_chroma, pitch_weaved_chroma);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_chroma, 0, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetAddressMode(cu_tr_chroma, 1, CU_TR_ADDRESS_MODE_CLAMP);
    CHECK_CU_RESULT
    cr = cuTexRefSetFilterMode(cu_tr_chroma, CU_TR_FILTER_MODE_LINEAR);
    CHECK_CU_RESULT
    cr = cuTexRefSetFlags(cu_tr_chroma, CU_TRSF_NORMALIZED_COORDINATES);
    CHECK_CU_RESULT
    cr = cuParamSetTexRef(cu_func_csc, CU_PARAM_TR_DEFAULT, cu_tr_chroma);
    CHECK_CU_RESULT
    cr = cuFuncSetBlockShape(cu_func_csc, 16, 16, 1);
    CHECK_CU_RESULT
    cr = cuLaunchGrid(cu_func_csc, (trace_stream_conf_ptr->out_width + 15) / 16, (trace_stream_conf_ptr->out_height + 15)/ 16);
    CHECK_CU_RESULT

    // Copy results back to VDPAU surface

    memset(&copy, 0, sizeof(copy));
    copy.srcXInBytes = 0;
    copy.srcY = 0;
    copy.srcMemoryType = CU_MEMORYTYPE_DEVICE;
    copy.srcDevice = cu_dptr_out;
    copy.srcPitch = pitch_out;
    copy.dstXInBytes = 0;
    copy.dstY = 0;
    copy.dstMemoryType = CU_MEMORYTYPE_ARRAY;
    copy.dstArray = aro;
    copy.WidthInBytes = trace_stream_conf_ptr->out_width * 4;
    copy.Height = trace_stream_conf_ptr->out_height;
    cr = cuMemcpy2D(&copy);
    CHECK_CU_RESULT

    if (opt_gold_surface_type == SURFACE_TYPE_CUDA_DRIVER_INTEROP_OUT) {
        cuda_driver_do_gold_interop_out(frame_number, aro);
    }

    // Unmap

    cr = cuGraphicsUnmapResources(1, &ros->gr, 0);
    CHECK_CU_RESULT
    cr = cuGraphicsUnmapResources(1, &rvs->gr, 0);
    CHECK_CU_RESULT
}

Traceplay_Callbacks tpcb = {
    traceplay_video_surface_post_create,
    traceplay_output_surface_post_create,
    traceplay_video_surface_pre_destroy,
    traceplay_output_surface_pre_destroy,
    traceplay_display_video_surface,
    traceplay_wait_for_output_surface_idle,
    traceplay_display_output_surface,
    traceplay_process_surface
};

void initX11(void)
{
    unsigned long windowMask;
    XSetWindowAttributes winAttrib;

    // X11 setup

    x_display = XOpenDisplay(NULL);
    if (!x_display) {
        printf("ERROR at %s:%d: XOpenDisplay failed\n", __FILE__, __LINE__);
        common_exit_error();
    }
    x_screen = DefaultScreen(x_display);

    windowMask = CWBorderPixel; 
    winAttrib.border_pixel = BlackPixel(x_display, x_screen);
    winAttrib.override_redirect = 0;

    x_window = XCreateWindow(
        x_display,
        RootWindow(x_display, x_screen),
        100,
        100,
        trace_stream_conf_ptr->out_width,
        trace_stream_conf_ptr->out_height,
        0,
        CopyFromParent,
        InputOutput,
        CopyFromParent,
        windowMask,
        &winAttrib
    );

    XSelectInput(x_display, x_window, StructureNotifyMask);

    XMapRaised(x_display, x_window);

    XFlush(x_display);
}

void finiX11(void)
{
    XDestroyWindow(x_display, x_window);
    XCloseDisplay(x_display);
}

void initCUDA(void)
{
    CUresult cr;

    cr = cuInit(0);
    CHECK_CU_RESULT

    cr = cuVDPAUGetDevice(&cu_dev, vdp_device, vdp_get_proc_address);
    CHECK_CU_RESULT

    cr = cuVDPAUCtxCreate(&cu_ctx, CU_CTX_BLOCKING_SYNC, cu_dev, vdp_device, vdp_get_proc_address);
    CHECK_CU_RESULT

    char *module_path = getenv("TRACEPLAY_CUDA_DRIVER_MODULE_PATH");
    if (!module_path) {
        module_path = ".";
    }
    char module_filename[FILENAME_MAX];
    snprintf(module_filename, sizeof(module_filename), "%s/%s", module_path, "traceplay_cuda_kernels.ptx");
    module_filename[sizeof(module_filename) - 1] = '\0';

    cr = cuModuleLoad(&cu_mod_vmr, module_filename);
    CHECK_CU_RESULT
    cr = cuModuleGetFunction(&cu_func_luma_weave, cu_mod_vmr, "luma_weave");
    CHECK_CU_RESULT
    cr = cuModuleGetFunction(&cu_func_chroma_weave, cu_mod_vmr, "chroma_weave");
    CHECK_CU_RESULT
    cr = cuModuleGetFunction(&cu_func_csc, cu_mod_vmr, "csc");
    CHECK_CU_RESULT

    cr = cuModuleGetTexRef(&cu_tr_luma_top, cu_mod_vmr, "luma_top");
    CHECK_CU_RESULT
    cr = cuModuleGetTexRef(&cu_tr_luma_bottom, cu_mod_vmr, "luma_bottom");
    CHECK_CU_RESULT
    cr = cuModuleGetTexRef(&cu_tr_chroma_top, cu_mod_vmr, "chroma_top");
    CHECK_CU_RESULT
    cr = cuModuleGetTexRef(&cu_tr_chroma_bottom, cu_mod_vmr, "chroma_bottom");
    CHECK_CU_RESULT
    cr = cuModuleGetTexRef(&cu_tr_luma, cu_mod_vmr, "luma");
    CHECK_CU_RESULT
    cr = cuModuleGetTexRef(&cu_tr_chroma, cu_mod_vmr, "chroma");
    CHECK_CU_RESULT

    cr = cuMemAllocPitch(&cu_dptr_out, &pitch_out, trace_stream_conf_ptr->out_width * 4, trace_stream_conf_ptr->out_height, 4);
    CHECK_CU_RESULT
    cr = cuMemAllocPitch(&cu_dptr_weaved_luma, &pitch_weaved_luma, trace_stream_conf_ptr->vid_width, trace_stream_conf_ptr->vid_height, 4);
    CHECK_CU_RESULT
    cr = cuMemAllocPitch(&cu_dptr_weaved_chroma, &pitch_weaved_chroma, trace_stream_conf_ptr->vid_width, (trace_stream_conf_ptr->vid_height / 2), 4);
    CHECK_CU_RESULT
}

void finiCUDA(void)
{
    CUresult cr;

    if (cu_dptr_weaved_chroma) {
        cr = cuMemFree(cu_dptr_weaved_chroma);
        CHECK_CU_RESULT
        cu_dptr_weaved_chroma = 0;
    }

    if (cu_dptr_weaved_luma) {
        cr = cuMemFree(cu_dptr_weaved_luma);
        CHECK_CU_RESULT
        cu_dptr_weaved_luma = 0;
    }

    if (cu_dptr_out) {
        cr = cuMemFree(cu_dptr_out);
        CHECK_CU_RESULT
        cu_dptr_out = 0;
    }

    if (cu_mod_vmr) {
        cr = cuModuleUnload(cu_mod_vmr);
        CHECK_CU_RESULT
        cu_mod_vmr = 0;
    }

    if (cu_ctx) {
        cr = cuCtxDestroy(cu_ctx);
        CHECK_CU_RESULT
        cu_ctx = 0;
    }
}

void initVDPAU(void)
{
    VdpStatus vdp_st;

    // Create Device

    vdp_st = vdp_device_create_x11(
        x_display,
        x_screen,
        &vdp_device,
        &vdp_get_proc_address
    );
    CHECK_ST

    initCUDA();

    // Entry point retrieval

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_DEVICE_DESTROY,
        (void **)&vdp_device_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_TARGET_DESTROY,
        (void **)&vdp_presentation_queue_target_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_CREATE,
        (void **)&vdp_presentation_queue_create
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_DESTROY,
        (void **)&vdp_presentation_queue_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_DISPLAY,
        (void **)&vdp_presentation_queue_display
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_BLOCK_UNTIL_SURFACE_IDLE,
        (void **)&vdp_presentation_queue_block_until_surface_idle
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_TARGET_CREATE_X11,
        (void **)&vdp_presentation_queue_target_create_x11
    );
    CHECK_ST

    vdp_st = vdp_presentation_queue_target_create_x11(
        vdp_device,
        x_window,
        &vdp_flip_target
    );
    CHECK_ST

    vdp_st = vdp_presentation_queue_create(
        vdp_device,
        vdp_flip_target,
        &vdp_flip_queue
    );
    CHECK_ST
}

void finiVDPAU(void)
{
    vdp_presentation_queue_destroy(
        vdp_flip_queue
    );

    vdp_presentation_queue_target_destroy(
        vdp_flip_target
    );

    vdp_device_destroy(
        vdp_device
    );
}

void help()
{
    printf("app [options]\n");
    common_print_help();
}

int main(int argc, char * * argv)
{
    int ret;

    app_name = "vdpau_traceplay_cuda_driver";
    support_opt_gold_surface_type_mask =
        (1 << SURFACE_TYPE_VDP_VID_DEC_ORDER) |
        (1 << SURFACE_TYPE_VDP_VID_DISP_ORDER) |
        (1 << SURFACE_TYPE_VDP_OUT) |
        (1 << SURFACE_TYPE_CUDA_DRIVER_INTEROP_VID) |
        (1 << SURFACE_TYPE_CUDA_DRIVER_INTEROP_OUT);

    common_init();

    argc--;
    argv++;
    while (argc) {
        if ((ret = common_parse_arg(&argc, &argv)) != COMMON_PARSE_ARG_NOT_PARSED) {
            if (ret != COMMON_PARSE_ARG_OK) {
                help();
                common_exit_error();
            }
        }
        else {
            printf("ERROR: Unknown option '%s'\n", argv[0]);
            help();
            common_exit_error();
        }
        argc--;
        argv++;
    }

    if (common_check_args() != COMMON_CHECK_ARGS_OK) {
        help();
        common_exit_error();
    }

    common_load_stream();

    initX11();
    initVDPAU();
    common_playback_init(
        vdp_device,
        vdp_get_proc_address,
        (opt_gold_surface_type == SURFACE_TYPE_CUDA_DRIVER_INTEROP_VID)
    );
    struct Traceplay_Ctx *traceplay_ctx = traceplay_init(
        vdp_device,
        vdp_get_proc_address,
        TRACEPLAY_PLAY_TYPE_CUSTOM_MIXER,
        &tpcb,
        trace_stream_conf_ptr
    );
    CHECK_PTR(traceplay_ctx)
    traceplay_play(
        traceplay_ctx,
        trace_cmds_ptr,
        opt_repeat_count
    );
    traceplay_fini(
        traceplay_ctx
    );
    common_playback_fini();
    finiCUDA();
    finiVDPAU();
    finiX11();

    common_exit_ok();
    return 0;
}

